from .exceptions import InvalidFlag


class Flags:

    escape_ansi_code = '\x1b['
    escape_ansi_end = 'm'

    #### COLORS (FG CODES)
    BLACK = 'k', '30'
    RED = 'r', '31'
    GREEN = 'n', '32'
    YELLOW = 'y', '33'
    BLUE = 'b', '34'
    MAGENTA = 'm', '35'
    CYAN = 'c', '36'
    WHITE = 'w', '37'
    GREY = 'g', '90'

    DEFAULT = 'p', '10'

    #### FORMATS
    BOLD = 'B', '1'
    DIM = 'D', '2'
    ITALIC = 'I', '3'
    UNDERLINE = 'U', '4'
    HIGHLIGHT = 'H', '7'
    STRIKE = 'S', '9'

    #### END OF LINE
    reset = '0'

    @classmethod
    def get_end_of_line(cls):
        """ Defined method to get the 'reset' code """
        return cls.escape_ansi_code + cls.reset + cls.escape_ansi_end

    @classmethod
    def join_flags(cls, flags):
        """ Given a set of flags, returned the final ansi code to add to the text"""
        return "%s%s%s" % (cls.escape_ansi_code, ';'.join(flags), cls.escape_ansi_end)

    @classmethod
    def get_flags(cls):
        """
        returns a dictionary where the flag is the key and the attribute
        name is the value
        """
        return {y[0]: x for x, y in vars(cls).items() if x.isupper() and isinstance(y, tuple) and len(y[0]) == 1}

    @classmethod
    def get_flag_values(cls, flags):
        """ returns a list of the escaped values for the flag labels """
        available_flags = cls.get_flags()
        flags_values = []
        for flag in flags.replace(' ', ''):
            if flag not in available_flags:
                raise InvalidFlag(flag)
            else:
                if hasattr(cls, available_flags[flag]):
                    flags_values.append(getattr(cls, available_flags[flag])[1])
        return flags_values




import platform

from .exceptions import (
    InvalidInputType, BoolOptionsNotValid,
    IntOptionsNotValid, FloatOptionsNotValid
)
from .flags import Flags

LINUX = 'Linux'
WINDOWS = 'Windows'
OSX = 'Darwin'

# For format() and format_input()
default_end = '\n'


class Printy:
    """
    Applies a format to the output of the print statement according
    to the flag (or flags).

    We can either set a global set of flags like >>> printy('Some text', 'rB')
    or set inline formats with the especial characters
    like >>> printy('[rB]Some@ [y]text@')
    """

    # For inline formatting we'll use special characters to catch the flags
    end_format_char = '@'
    open_flag_char = '['
    close_flag_char = ']'
    special_chars = [end_format_char, open_flag_char, close_flag_char]

    # Actions for inline formats
    START_FLAGS = 'start_flags'
    START_FORMAT = 'start_format'
    END_FORMAT = 'end_format'
    ESCAPE_CHAR = 'escape_char'

    def __init__(self):
        self.platform = platform.system()
        self.virtual_terminal_processing = self.set_windows_console_mode()

    def set_windows_console_mode(self):
        """
        For Windows os to work and get the escape sequences correctly,
        we'll need to enable the variable ENABLE_VIRTUAL_TERMINAL_PROCESSING

        @Thanks to Mihir Singh (mihirs16) for this big improvement
        """
        # In case there is some error while setting it up, we returns False to
        # indicate that windows will not print the escape sequences correctly,
        # so we can print out the cleaned text
        if self.platform == WINDOWS:
            try:
                from ctypes import windll
                k = windll.kernel32
                k.SetConsoleMode(k.GetStdHandle(-11), 7)
                return True
            except ImportError:
                return False
        return False

    @classmethod
    def _define_char(cls, prev, current):
        """
        Helper method that'll tell us if a character has to be treated as a
        special one or it is part of the text that 's intended to be printed
        out.

        Takes the previous character and the current character, in case it is
        one of the special characters defined in the class and is prepended by
        a '\' means that it has not to be treated as a special one

        Returns True if it's a special character
        """
        if current in cls.special_chars:
            if prev != '\\':
                return True
        return False

    @classmethod
    def _check_special_char_position(cls, last_special, special):
        """
        Returns an action to execute if the character is well placed. It should
        only be applied over special characters.

        If it's not well placed, the character will be included in the text
        """
        if special == cls.open_flag_char:
            # In this case the 'last_special' must always be
            # the 'end_format_char' or None if it's the first appearing
            if last_special not in [cls.end_format_char, None]:
                return cls.ESCAPE_CHAR
            else:
                return cls.START_FLAGS
        elif special == cls.close_flag_char:
            # In this case the 'last_special' must always be the 'open_flag_char'
            if last_special != cls.open_flag_char:
                return cls.ESCAPE_CHAR
            else:
                return cls.START_FORMAT
        elif special == cls.end_format_char:
            # In this case the 'last_special' must always be the 'close_flag_char'
            # Or None if the text does not include any other formatting character
            if last_special in [cls.open_flag_char, cls.end_format_char, None]:
                return cls.ESCAPE_CHAR
            else:
                return cls.END_FORMAT

    @classmethod
    def _replace_escaped(cls, text):
        """ Replaces escaped special characters for the character itself """
        for special_char in cls.special_chars:
            text = text.replace('\\' + special_char, special_char)
        return text

    @classmethod
    def _get_inline_format_as_tuple(cls, text):
        """
        In case some inline formats have been applied we need to get a list of
        tuples indicating the formats to be applied via flags and the text
        where the format should be applied, for instance, if the text is:
        "[rB]Some@ Te[H]xt@"
        We'll get the list [('Some', 'rB), (' Te', None), ('xt', 'H')]
        """
        prev = ''  # Stores the last character in the loop
        last_special_char = None
        list_of_formats = []  # Final list to be returned

        # While looping the text, we'll get the formats to be applied to
        # certain section of that text
        section_text = []
        section_flags = []
        current_action = cls.START_FORMAT

        # Will tell us when we're at the last character of the loop
        counter = 0
        # Will tell us if we need to reset the sections variables
        close_section = False
        for char in text:
            is_special = cls._define_char(prev, char)
            if is_special:
                action = cls._check_special_char_position(last_special_char, char)

                if action == cls.ESCAPE_CHAR:
                    # Add the character to the text
                    if current_action == cls.START_FLAGS:
                        section_flags.append(char)
                    elif current_action == cls.START_FORMAT:
                        section_text.append(char)
                else:
                    # Here we know that the special character is well placed
                    # and has a special meaning
                    current_action = action

                    if current_action == cls.END_FORMAT:
                        current_action = cls.START_FORMAT
                        close_section = True
                    if last_special_char in [cls.end_format_char, None]:
                        # Here we'll catch 'open_flag_char's
                        current_action = cls.START_FLAGS
                        close_section = True
                    last_special_char = char
            else:
                if current_action == cls.START_FLAGS:
                    section_flags.append(char)
                elif current_action == cls.START_FORMAT:
                    section_text.append(char)
            prev = char
            counter += 1

            if counter == len(text) or close_section:
                # Reset the 'section_*' lists and add them (joined)
                # to the final list
                list_of_formats.append((
                    ''.join(section_text),
                    ''.join(section_flags) if len(section_flags) > 0 else None
                ))
                section_text = []
                section_flags = []
                close_section = False

        return list_of_formats

    @classmethod
    def _get_cleaned_text(cls, text):
        """ Returns the cleaned value, with no formats """
        tuple_text = cls._get_inline_format_as_tuple(text)
        return cls._replace_escaped(''.join(x[0] for x in tuple_text))

    def get_formatted_text(self, value, flags=None, predefined=None):
        """
        Applies the format specified by the 'flags' to the 'value'.

        If 'flag's is passed, 'predefined' will be omitted.
        """
        if self.platform == WINDOWS and not self.virtual_terminal_processing:
            text = self._get_cleaned_text(value)
        else:
            if flags:
                flags_values = Flags.get_flag_values(flags)
                value = self._get_cleaned_text(value)
                text = "%s%s%s" % (
                    Flags.join_flags(flags_values),
                    value,
                    Flags.get_end_of_line()
                )
            else:
                tuple_text = self._get_inline_format_as_tuple(value)
                text = ''
                for section in tuple_text:
                    section_text = self._replace_escaped(section[0])
                    section_flags = section[1] or predefined
                    if section_flags:
                        flags_values = Flags.get_flag_values(section_flags)
                        text += "%s%s%s" % (
                            Flags.join_flags(flags_values),
                            section_text,
                            Flags.get_end_of_line()
                        )
                    else:
                        text += section_text
        return text

    @staticmethod
    def read_file(file):
        """ Given a file path, we read it and print it out """
        file = str(file)
        with open(file) as f:
            text = f.read()
        return text

    def format(self, value='', flags=None, predefined=None, file='', end=default_end):
        """ Prints out the value """
        value = self.read_file(file) if file else value
        print(self.get_formatted_text(value, flags, predefined), end=end)

    ##### Inputy

    # Types (str is the default)
    BOOL = 'bool'
    INT = 'int'
    FLOAT = 'float'
    STR = 'str'
    types = [BOOL, INT, FLOAT, STR]

    @staticmethod
    def construct_end_line(*args, **kwargs):
        """
        For inputy, given certain parameters we can create a probable end of
        line, this is, a helper message for the user, in case we set, for
        example, if we set that the value must be an integer, we could add
        that information to the end user.
        """
        end = kwargs.get('end', default_end)
        value_type = kwargs.get('end', None)
        pass

    @staticmethod
    def get_bool_options(bool_options):
        """
        Strips the string passed as bool_options and returns the valid values,
        let's say the bool_options = 'i{Y/n}', then it will return a tuple
        containing:
            insensitive, true, false = True, 'Y', 'n'
        """
        insensitive, true_value, false_value = True, True, False
        if bool_options:
            opts = list(bool_options.replace('{', '').replace('}', '').replace('/', ''))
            # It might end up with 2 or 3 values, if its 3, the first one must be
            # the case insensitive indicator 'i'
            if len(opts) == 3 and opts[0] != 'i':
                raise BoolOptionsNotValid(bool_options)
            elif len(opts) > 3 or len(opts) <= 1:
                raise BoolOptionsNotValid(bool_options)
            elif len(opts) == 3:
                insensitive, true_value, false_value = True, opts[1], opts[2]
            else:
                true_value, false_value = opts
                insensitive = False

        return insensitive, true_value, false_value

    @staticmethod
    def get_int_options(int_options):
        """ Checks that only '+' or '-' are passed """
        # the default is None, which will accept any number, positive or negative
        if int_options:
            if len(int_options) > 1:
                raise IntOptionsNotValid(int_options)
            else:
                if int_options in ('+', '-'):
                    return int_options
                else:
                    raise IntOptionsNotValid(int_options)
        return None

    def check_boolean(self, value, bool_options=None):
        """
        Validates the value when the type must be a boolean, returns a boolean
        specifying whether it is a valid value, and if it is, returns the final
        value (after conversions if necessary)
        """
        insensitive, true_value, false_value = self.get_bool_options(bool_options)
        true_value = str(true_value)
        false_value = str(false_value)
        error_msg = "%s is not a valid value, enter %s or %s" % (
            value,
            true_value,
            false_value,
        )
        if insensitive:
            value = value.lower()
            true_value = true_value.lower()
            false_value = false_value.lower()
        if value == true_value:
            return True, True
        elif value == false_value:
            return False, True
        else:
            self.format(error_msg)
            return False, False

    def check_integer(self, value, int_options=None):
        """
        Validates the value when the type must be an integer, returns a boolean
        specifying whether it is a valid value, and if it is, returns the final
        value (after conversions if necessary)
        """
        # the only options allowed for integer types are '+' and '-'
        opt = self.get_int_options(int_options)
        error_msg = (
            "%s is not a valid number,please enter a [b]rounded@"
            " number, please check you are not adding some "
            "decimal digits" % value
        )
        if opt:
            error_msg += ', make sure also it is a %s number' % (
                'positive' if opt == '+' else 'negative'
            )
        # Let's try to convert it to integer
        valid_value = False
        try:
            value = int(value)
        except (ValueError, TypeError):
            self.format(error_msg)
        else:
            if opt:
                if opt == '+' and value >= 0:
                    valid_value = True
                elif opt == '-' and value < 0:
                    valid_value = True
                else:
                    valid_value = False
            else:
                valid_value = True

        return value, valid_value

    def check_float(self, value, float_options=None):
        """
        Validates the value when the type must be an float, similar to integer check,
        but now it can have decimal digits, returns a boolean specifying whether it
        is a valid value, and if it is, returns the final
        value (after conversions if necessary)
        """
        # Same integer options are allowed
        opt = self.get_int_options(float_options)
        error_msg = (
            "%s is not a valid number" % value
        )
        if opt:
            error_msg += ', make sure also it is a %s number' % (
                'positive' if opt == '+' else 'negative'
            )
        # Let's try to convert it to integer
        valid_value = False
        try:
            value = float(value)
        except (ValueError, TypeError):
            self.format(error_msg)
        else:
            if opt:
                if opt == '+' and value >= 0:
                    valid_value = True
                elif opt == '-' and value < 0:
                    valid_value = True
                else:
                    valid_value = False
            else:
                valid_value = True

        return value, valid_value

    def format_input(self, *args, **kwargs):
        """
        Colorize the text prompted by input().

        Also, it takes an additional parameter 'type', to tell the prompt not
        to accept a format other than the specified. As every input is converted
        to strings, a string that can be converted to the specified type is
        allowed. For example, if type=int, then the user would be forced to
        enter a number or a string that can be converted into an integer
        """

        # If passed, we'll force the user to write a value with the specific
        # input_types' format.
        input_type = kwargs.get('type', self.STR)
        if input_type not in self.types:
            raise InvalidInputType(input_type)

        # bool_options will override the default True/False
        # it has to be passed as a string with the following syntax:
        # bool_options="i{True/False}", where 'i', if passed, will tell us
        # that the values are case insensitive, otherwise, values will be forced
        # to a exact match
        options = kwargs.get('options', '')
        # when no value is entered, the default will be added
        default = kwargs.get('default', '')

        # remove extra parameters so we don't pass it to
        # the get_formatted_text function
        if 'type' in kwargs:
            kwargs.pop('type')
        if 'options' in kwargs:
            kwargs.pop('options')

        # Include the value for the get_formatted_text function
        if len(args) == 0:
            args = ['']

        # Will tell us whether the user sent a value value according to the
        # specified type or not
        valid_value = False
        result = None
        while not valid_value:
            # Prints out the message if any was passed
            result = str(input(self.get_formatted_text(*args, **kwargs)))

            if result == '':
                result = default

            if input_type == self.BOOL:
                # now let's try to convert the value to a Boolean
                result, valid_value = self.check_boolean(result, options)

            elif input_type == self.INT:
                # Let's try to convert it to integer
                result, valid_value = self.check_integer(result, options)

            elif input_type == self.FLOAT:
                # Almost the same for integer, but this time it just have to
                # be a number, rounded or not
                result, valid_value = self.check_float(result, options)
            else:
                result, valid_value = str(result), True

        return result



import unittest
from unittest import mock
from printy.exceptions import (
    InvalidFlag, InvalidInputType, BoolOptionsNotValid,
    IntOptionsNotValid, FloatOptionsNotValid
)
from printy.core import Printy, WINDOWS
from printy.flags import Flags


class TestGlobalFlagsPrinty(unittest.TestCase):
    """ Test case for formatting with a global set of flags specified """

    def setUp(self):
        self.sample_text = "Some Text To Print Out"
        self.printy = Printy()
        self.raw_text = self.printy.get_formatted_text

    def test_empty_value(self):
        """ Tests passing an empty value print's nothing"""
        text = ''
        result = self.raw_text(text)

        self.assertEqual(result, text)

    def test_empty_value_with_flags(self):
        """
        Tests that passing and empty value with some flags returns the
        escape ansi characters
        """
        text = ''
        flags = 'rBH'
        result = self.raw_text(text, flags)
        expected_result = "%s%s" % (
                Flags.join_flags(Flags.get_flag_values(flags)),
                Flags.get_end_of_line()
        )

        self.assertEqual(result, expected_result)

    def test_single_invalid_flag(self):
        """
        Tests that passing an invalid flag (only one)
        raises and exception
        """
        invalid_flag = 'P'
        with self.assertRaises(InvalidFlag):
            self.printy.format(self.sample_text, invalid_flag)

    def test_multiple_invalid_flag(self):
        """
        Tests that passing multiple invalid flags raises an
        exception with the first invalid flag found
        """
        # P and G are invalid, should raise InvalidFlag
        # with 'P' as invalid flag
        flags = 'yBPGr'
        with self.assertRaises(InvalidFlag) as e:
            self.printy.format(self.sample_text, flags)
        self.assertEqual(e.exception.flag, 'P')

    def tests_always_closing_format(self):
        """
        Tests that the returned text is always ended with the closing tag
        """
        result = self.raw_text(self.sample_text, 'r')
        closing_tag = result[-4:]
        self.assertEqual(len(closing_tag), 4)
        self.assertEqual(closing_tag, Flags.get_end_of_line())

    def test_no_flag_parameter_passed(self):
        """
        Tests that passing no flag parameter will return a default value
        """
        result = self.raw_text(self.sample_text)
        self.assertEqual(result, self.sample_text)

    def test_empty_flag(self):
        """
        Test that passing and empty string as a flag still returns the
        default value
        """
        result = self.raw_text(self.sample_text, '')
        self.assertEqual(result, self.sample_text)

    def test_flags_with_spaces_in_between(self):
        """
        Tests that passing a set of flags with some spaces in between
        (like 'yB H U') still applies the desired formats
        """
        desired_flags = 'yBH'
        flags_with_one_space = 'yB H'
        flags_with_multiple_spaces = 'y B H'
        result_one = self.raw_text(self.sample_text, desired_flags)
        result_two = self.raw_text(self.sample_text, flags_with_one_space)
        result_three = self.raw_text(self.sample_text, flags_with_multiple_spaces)

        self.assertTrue(result_one == result_two == result_three)

    @mock.patch('printy.core.Printy.set_windows_console_mode', return_value=True)
    def test_virtual_terminal_processing_on_windows(self, mock_console_mode):
        """
        Tests that if platform is windows, then then returns True
        """
        self.printy.platform = WINDOWS
        virtual_terminal_processing = mock_console_mode()

        self.assertTrue(virtual_terminal_processing)

    def test_return_cleaned_value_if_windows_is_not_properly_configured(self):
        """
        Tests that if printy virtual_console_mode is false, then it returns the
        cleaned_text
        """
        flags = 'rBH'
        # Changes platform to Windows
        self.printy.platform = WINDOWS
        self.printy.virtual_terminal_processing = False

        result_one = self.raw_text(self.sample_text, flags)

        self.assertEqual(result_one, self.sample_text)


class TestInlineFlagsPrinty(unittest.TestCase):
    """ Test case for inline formatting """

    def setUp(self):
        self.printy = Printy()
        self.raw_text = self.printy.get_formatted_text

    def test_inline_format_with_global_flags(self):
        """
        Tests that passing a text with inline formatting and also a global
        set of flags takes this last one as the format to be applied
        """
        inline_formatted = "[y]Hey you@"
        no_format = 'Hey you'
        global_flags = 'rB'
        result_one = self.raw_text(inline_formatted, global_flags)
        result_two = self.raw_text(no_format, global_flags)

        self.assertEqual(result_one, result_two)

    def test_inline_format_without_ending_format_character(self):
        """
        Tests that passing an inline formatted text without the ending
        formatting character still returns the formatted text
        """
        result_one = self.raw_text('[y]Hey you')
        result_two = self.raw_text('[y]Hey you@')

        self.assertEqual(result_one, result_two)

    def test_escape_special_characters(self):
        """ Tests that escaping special characters prints them out """
        inline_text_one = '[y]myemail\@mydomain.com@'
        global_text_one = 'myemail@mydomain.com', 'y'

        inline_text_two = '[bH]Some text \@@'
        global_text_two = 'Some text @', 'bH'

        inline_result_one = self.raw_text(inline_text_one)
        global_result_one = self.raw_text(global_text_one[0], global_text_one[1])

        inline_result_two = self.raw_text(inline_text_two)
        global_result_two = self.raw_text(global_text_two[0], global_text_two[1])

        self.assertEqual(inline_result_one, global_result_one)
        self.assertEqual(inline_result_two, global_result_two)

    def test_multiple_sections(self):
        """ Test that formats are applied correctly to each section """

        section_one = "Some"
        section_two = ' '
        section_three = 'text'
        global_format_one = self.raw_text(section_one, 'rB')
        global_format_two = self.raw_text(section_two)
        global_format_three = self.raw_text(section_three, 'y')
        joined_global_format = global_format_one + global_format_two + global_format_three

        inline_text = '[rB]Some@ [y]text@'
        inline_format = self.raw_text(inline_text)

        self.assertEqual(inline_format, joined_global_format)

    def test_read_file(self):
        """ Test retrieving the text from a file """
        text_in_file = 'printy'
        file_name = 'printy_file'
        with mock.patch('builtins.open', mock.mock_open(read_data=text_in_file)) as m:
            result = self.printy.read_file(file_name)

        m.assert_called_once_with(file_name)
        self.assertEqual(result, text_in_file)


class TestInputy(unittest.TestCase):
    """
    Test case for inputy functionality

    Here, it is not necessary to test whether the prompted message has the
    correct format because it uses the methods already tested in the Printy
    test cases
    """

    def setUp(self):
        self.inputy = Printy()

    str_valid_test = "Valid String"
    int_valid_test = 23
    float_valid_test = 45.6
    bool_valid_test = False

    def test_get_bool_options_case_insensitive(self):
        """
        Tests returning True for the insensitive value if 'i' is passed as
        the first character of the 'options' parameter
        """
        options = 'i{y/n}'
        insensitive, true_option, false_option = self.inputy.get_bool_options(options)

        self.assertTrue(insensitive)
        self.assertEqual(true_option, 'y')
        self.assertEqual(false_option, 'n')

    def test_get_bool_options_case_sensitive(self):
        """
        Tests returning True for the insensitive value if no 'i' is passed as
        the first character of the 'options' parameter
        """
        options = 'y/n'
        insensitive, true_option, false_option = self.inputy.get_bool_options(options)

        self.assertFalse(insensitive)
        self.assertEqual(true_option, 'y')
        self.assertEqual(false_option, 'n')

    def test_default_value_options(self):
        """
        Tests return the default 'True' and 'False' and case insensitive if no
        options are passed
        """
        options = ''
        insensitive, true_option, false_option = self.inputy.get_bool_options(options)

        self.assertTrue(insensitive)
        self.assertTrue(true_option)
        self.assertFalse(false_option)

    def test_passing_wrong_flag_for_case_insensitive(self):
        """
        Tests that passing a value different than 'i' for the insensitive
        flag in in 'options' parameter when the type is bool, raises an exception
        """
        options = 'f{y/n}'
        with self.assertRaises(BoolOptionsNotValid) as e:
            self.inputy.get_bool_options(options)
        self.assertEqual(e.exception.options, options)

    def test_passing_invalid_options_to_bool_type(self):
        """
        Tests that passing an invalid option value to the parameter 'options'
        raises an BoolOptionsNotValid
        """
        invalid_options = "jgnlf"
        with self.assertRaises(BoolOptionsNotValid) as e:
            self.inputy.get_bool_options(invalid_options)
        self.assertEqual(e.exception.options, invalid_options)

    def test_get_int_options(self):
        """ Test returning valid options"""
        positive = '+'
        negative = '-'
        int_options_positive = self.inputy.get_int_options(positive)
        int_options_negative = self.inputy.get_int_options(negative)

        self.assertEqual(positive, int_options_positive)
        self.assertEqual(negative, int_options_negative)

    def test_passing_long_character_as_int_options(self):
        """
        Tests that passing more than one character (only + or - are allowed)
        raises an exception
        """
        long_int_options_one = '+-'
        long_int_options_two = 'plus'

        with self.assertRaises(IntOptionsNotValid) as e_one:
            self.inputy.get_int_options(long_int_options_one)
        with self.assertRaises(IntOptionsNotValid) as e_two:
            self.inputy.get_int_options(long_int_options_two)
        self.assertEqual(e_one.exception.options, long_int_options_one)
        self.assertEqual(e_two.exception.options, long_int_options_two)

    def test_check_boolean_case_sensitive_returns_value_converted(self):
        """
        tests that passing a value (according to the options) to a type='bool'
        returns a converted value (True or False)
        """
        options = 'y/n'
        value_true = 'y'
        value_false = 'n'
        returned_value_true, valid_value_true = self.inputy.check_boolean(value_true, options)
        returned_value_false, valid_value_false = self.inputy.check_boolean(value_false, options)

        self.assertEqual(returned_value_true, True)
        self.assertEqual(returned_value_false, False)
        self.assertTrue(valid_value_true)
        self.assertTrue(valid_value_false)

    def test_check_boolean_case_insensitive_returns_value_converted(self):
        """
        tests that passing a value with different case (according to the options
        as case insensitive) to a type='bool' returns a converted value (True or False)
        """
        options = 'i{y/n}'
        value_true = 'Y'
        value_false = 'N'
        returned_value_true, valid_value_true = self.inputy.check_boolean(value_true, options)
        returned_value_false, valid_value_false = self.inputy.check_boolean(value_false, options)

        self.assertEqual(returned_value_true, True)
        self.assertEqual(returned_value_false, False)
        self.assertTrue(valid_value_true)
        self.assertTrue(valid_value_false)

    def test_check_boolean_invalid_value(self):
        """
        tests passing an invalid value to check_bool returns False as the value
        and also False as 'valid_value'
        """
        options = 'i{y/n}'
        value_true = 'Yes'
        value_false = 'No'
        returned_value_true, valid_value_true = self.inputy.check_boolean(value_true, options)
        returned_value_false, valid_value_false = self.inputy.check_boolean(value_false, options)

        self.assertEqual(returned_value_true, False)
        self.assertEqual(returned_value_false, False)
        self.assertFalse(valid_value_true)
        self.assertFalse(valid_value_false)

    def test_check_integer_returns_converted_value(self):
        """ tests that check_integer returns the value converted as integer"""
        value = 34
        returned_value, valid_value = self.inputy.check_integer(value)

        self.assertTrue(isinstance(returned_value, int))
        self.assertEqual(returned_value, value)
        self.assertTrue(valid_value)

    def test_check_integer_with_positive_option(self):
        """
        Tests that passing '+' as options for type='int' returns the converted
        value and valid_value as False if it is not a positive number
        """
        opts_positive = '+'
        valid_int = 34
        invalid_int = -34
        return_valid_int, valid_value_valid_int = self.inputy.check_integer(valid_int, opts_positive)
        return_invalid_int, valid_value_invalid_int = self.inputy.check_integer(invalid_int, opts_positive)

        self.assertTrue(isinstance(return_valid_int, int))
        self.assertEqual(return_valid_int, valid_int)
        self.assertTrue(valid_value_valid_int)

        self.assertTrue(isinstance(return_invalid_int, int))
        self.assertEqual(return_invalid_int, invalid_int)
        self.assertFalse(valid_value_invalid_int)

    def test_check_integer_with_negative_option(self):
        """
        Tests that passing '-' as options for type='int' returns the converted
        value and valid_value as False if it is not a negative number
        """
        opts_positive = '-'
        valid_int = -34
        invalid_int = 34
        return_valid_int, valid_value_valid_int = self.inputy.check_integer(valid_int, opts_positive)
        return_invalid_int, valid_value_invalid_int = self.inputy.check_integer(invalid_int, opts_positive)

        self.assertTrue(isinstance(return_valid_int, int))
        self.assertEqual(return_valid_int, valid_int)
        self.assertTrue(valid_value_valid_int)

        self.assertTrue(isinstance(return_invalid_int, int))
        self.assertEqual(return_invalid_int, invalid_int)
        self.assertFalse(valid_value_invalid_int)

    def test_check_float_returns_converted_value(self):
        """ tests that check_float returns the value converted as float"""
        value = 34.5
        returned_value, valid_value = self.inputy.check_float(value)

        self.assertTrue(isinstance(returned_value, float))
        self.assertEqual(returned_value, value)
        self.assertTrue(valid_value)

    def test_check_float_with_positive_option(self):
        """
        Tests that passing '+' as options for type='float' returns the converted
        value and valid_value as False if it is not a positive number
        """
        opts_positive = '+'
        valid_int = 34.5
        invalid_int = -34.5
        return_valid_int, valid_value_valid_int = self.inputy.check_float(valid_int, opts_positive)
        return_invalid_int, valid_value_invalid_int = self.inputy.check_float(invalid_int, opts_positive)

        self.assertTrue(isinstance(return_valid_int, float))
        self.assertEqual(return_valid_int, valid_int)
        self.assertTrue(valid_value_valid_int)

        self.assertTrue(isinstance(return_invalid_int, float))
        self.assertEqual(return_invalid_int, invalid_int)
        self.assertFalse(valid_value_invalid_int)

    def test_check_float_with_negative_option(self):
        """
        Tests that passing '-' as options for type='float' returns the converted
        value and valid_value as False if it is not a negative number
        """
        opts_positive = '-'
        valid_int = -34.0
        invalid_int = 34.0
        return_valid_int, valid_value_valid_int = self.inputy.check_float(valid_int, opts_positive)
        return_invalid_int, valid_value_invalid_int = self.inputy.check_float(invalid_int, opts_positive)

        self.assertTrue(isinstance(return_valid_int, float))
        self.assertEqual(return_valid_int, valid_int)
        self.assertTrue(valid_value_valid_int)

        self.assertTrue(isinstance(return_invalid_int, float))
        self.assertEqual(return_invalid_int, invalid_int)
        self.assertFalse(valid_value_invalid_int)

    @mock.patch('builtins.input', return_value=str_valid_test)
    def test_passing_no_parameters_returns_a_value_str(self, mock_input):
        """ Testing 'inputy' as a normal 'input()' function """
        result_str = self.inputy.format_input()
        self.assertEqual(result_str, self.str_valid_test)

    @mock.patch('builtins.input', return_value=int_valid_test)
    def test_passing_no_parameters_returns_a_value_str_from_int(self,
                                                                mock_input):
        """ Testing 'inputy' as a normal 'input()' function """
        result_str_from_int = self.inputy.format_input()
        self.assertEqual(result_str_from_int, str(self.int_valid_test))

    @mock.patch('builtins.input', side_effect=[str_valid_test, bool_valid_test, float_valid_test, None, int_valid_test])
    def test_passed_invalid_when_requested_int(self, mock_input):
        """
        Test that, when specifying the users has to enter an integer,
        the message is prompted until a valid number is passed
        """
        result_valid_int = self.inputy.format_input(type='int')

        self.assertEqual(result_valid_int, self.int_valid_test)

    @mock.patch('builtins.input', side_effect=[None, str_valid_test, bool_valid_test, float_valid_test])
    def test_passed_invalid_when_requested_float(self, mock_input):
        """
        Test that, when specifying the users has to enter a number,
        the message is prompted until a valid number is passed
        """
        result_valid_int = self.inputy.format_input(type='float')

        self.assertEqual(result_valid_int, self.float_valid_test)

    @mock.patch('builtins.input', side_effect=[str_valid_test, None, int_valid_test, bool_valid_test])
    def test_passed_invalid_when_requested_boolean(self, mock_input):
        """
        Test that, when specifying the user has to enter a boolean
        the message is prompted until a boolean is passed
        """
        result_valid_boolean = self.inputy.format_input(type='bool')

        self.assertEqual(result_valid_boolean, self.bool_valid_test)

    @mock.patch('builtins.input', side_effect=[str_valid_test, None, int_valid_test, 'true'])
    def test_passed_invalid_when_requested_boolean_str(self, mock_input):
        """
        Test that, when specifying the user has to enter a boolean
        the message is prompted until a test case insensitive with the name of
        one of the boolean values is passed
        """
        result_valid_boolean = self.inputy.format_input(type='bool')

        self.assertEqual(result_valid_boolean, True)

    @mock.patch('builtins.input', return_value=str_valid_test)
    def test_passing_and_invalid_input_type(self, mock_input):
        """
        Tests that passing and invalid input type raises an InvalidInputType
        exception. We mock the input() just in case the tests reaches that section
        """
        invalid_input_type = 'not_int_nor_float'
        with self.assertRaises(InvalidInputType) as e:
            self.inputy.format_input(type=invalid_input_type)
        self.assertEqual(e.exception.input_type, invalid_input_type)



from .core import Printy

__version__ = "1.2.0"

printy_instance = Printy()

# Main function to extend print() functionality
printy = printy_instance.format

# If user just want to get the formatted text with the ANSII escape code
raw_format = printy_instance.get_formatted_text

# Main function to extend input() functionality
inputy = printy_instance.format_input




from .core import Printy

helpme = """#########################################################
#########################################################

[bBH]PRINTY@

Printy lets you colorize and apply some standard formats to your text with
an intuitive and friendly API based on flags to specify the formats. You can
either apply a global format or inline formats to specific parts of your text!
And yes, this whole tutorial is written with Printy!!

[gI]NOTE: If you are on windows, you may not see some of the formats due
to Windows console limitations@

[n]Let's get started!@

[B]COLORS@

[g]'g' -> Applies a grey color to the text@
[r]'r' -> Applies a red color to the text@
[n]'n' -> Applies a green color to the text@
[y]'y' -> Applies a yellow color to the text@
[b]'b' -> Applies a blue color to the text@
[m]'m' -> Applies a magenta color to the text@
[c]'c' -> Applies a cyan color to the text@
[k]'k' -> Applies a black color to the text@
[w]'w' -> Applies a white color to the text@
[p]'p' -> Applies the predefined color to the text@

[B]FORMATS@

[B]'B' -> Applies a bold font weight to the text@
[D]'D' -> Decrease the intensity, aka Dim@
[U]'U' -> Applies an underline to the text@
[I]'I' -> Applies an italic font type to the text@
[H]'H' -> Highlights the text@
[S]'S' -> crosses out the text, aka Strike@


[B]HOW TO USE IT?@

First import printy:

[n]    >>>@ [r]from@ printy [r]import@ printy

Now apply one or more of the available flags to the text like this:

[n]    >>>@ printy([c]'Some text to be formatted'@[r],@ [c]'b'@)
    [b]Some text to be formatted@

[n]    >>>@ printy([c]'Some text to be formatted'@[r],@ [c]'bHI'@)
    [bHI]Some text to be formatted@

Or as Inline format, use the \[\] to specify the flags, and the \@ to finish
the format for a specific section:

[n]    >>>@ printy([c]'\[r\]Red\@ Default Color \[yH\]Yellow Highlighted\@ Default Color'@)
    [r]Red@ Default Color [yH]Yellow Highlighted@ Default Color

You can always override the whole format with a global flag:

[n]    >>>@ printy([c]'\[r\]Red\@ Default Color \[yH\]Yellow Highlighted\@ Default Color'@[r],@ [c]'b'@)
    [b]Red Default Color Yellow Highlighted Default Color@

Or you can change only the default color:

[n]    >>>@ printy([c]'\[r\]Red\@ Default Color \[yH\]Yellow Highlighted\@ Default Color'@[r], default@=[c]'b'@)
    [r]Red@ [b]Default Color@ [yH]Yellow Highlighted@ [b]Default Color@

[n]    >>>@ printy([c]'\[r\]Red\@ Default Color \[yH\]Yellow Highlighted\@ Default Color'@[r], default@=[c]'nBIU'@)
    [r]Red@ [nBIU]Default Color@ [yH]Yellow Highlighted@ [nBIU]Default Color@

If you need to use one of the special characters ('\[', '\]', '\@'), simply escape them ensure
they show up, or, in case they are not placed in the form '\[flags\]Text\@'), you can include
them without escaping, the algorithm will treat them as regular characters:

[n]    >>>@ printy([c]'\[B\]\\\[myemail\\\@mydomain.com\\\]\@'@)
    [B]\[myemail\@mydomain.com\]@

[n]    >>>@ [g]# Here, the emails's '\@' would close the '\[B\]' statement,@
[n]    >>>@ [g]# so we need to escape it anyway. But look at how we don't @
[n]    >>>@ [g]# escape the \[ and \] characters. @
[n]    >>>@ printy([c]'\[B\]\[myemail\\\@mydomain.com\]\@'@)
    [B][myemail\@mydomain.com]@

Now you can do stuffs like:

[n]    >>>@ text = 'Hello world, python is awesome!!!'
[n]    >>>@ printy(text.replace([c]'python'@[r],@ [c]'\[r\]python\@'@))
    Hello world, [r]python@ is awesome

Or some html highlighting:

[n]    >>>@ html = (
[n]    ...@ [c]'<div \[p\]class=\@\[n\]"active"\@ \[p\]id=\@\[n\]"my-div"\@>'@
[n]    ...@ [c]'    <span \[p\]data-some-data=\@\[n\]"extra-data"\@>\[p\]Some text\@</span>'@
[n]    ...@ [c]'</div>'@)
[n]    >>>@ printy(html[r], default@=[c]'r'@)
    [r]<div@ class=[n]'active'@ id=[n]'my-div'@[r]>@
        [r]<span@ data-some-data=[n]'extra-data'@[r]>@Some text[r]</span>@
    [r]</div>@

Or, you can use it with python's formatting strings as well:

[n]    >>>@ minutes = [b]60@
[n]    >>>@ printy([c]f'A day has \[y\]@[r]{@[p]minutes * @[b]24@[r]}@[c]\@ minutes'@)
    A day has [y]1400@ minutes

[B]What about @[b]input@()?

printy comes with a wrapper for the python builtins input() function

[n]    >>>@ [r]from@ printy [r]import@ inputy

Try adding adding some formats the same way you did with printy.

[By]Check this link for a full documentation!@

https://github.com/edraobdu/printy    
"""

print(Printy().get_formatted_text(helpme))




import os

from setuptools import find_packages, setup

from printy import __version__

readme_path = os.path.join(os.path.dirname(__file__), "README.md")
with open(readme_path) as fh:
    long_description = fh.read()

setup(
    name="printy",
    version=__version__,
    url="https://github.com/edraobdu/printy",
    author="Edgardo Obregón",
    author_email="edraobdu@gmail.com",
    description="Colorize the print statement by global or inline flags",
    license="MIT",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    python_requires='>=3.5'
)



#### Printy


class InvalidFlag(Exception):
    """ raised when an invalid flag is passed to the 'printy' object"""

    def __init__(self, flag):
        self.flag = flag

    def __str__(self):
        return "'%s' is not a valid flag" % self.flag


#### Inputy


class InvalidInputType(Exception):
    """ raised when an invalid 'type' is passed to the 'inputy' function"""

    def __init__(self, input_type):
        self.input_type = input_type

    def __str__(self):
        return "'%s' is not a valid type" % self.input_type


class OptionsNotValid(Exception):
    """ Raises when the options for the inputy function is wrong written """

    def __init__(self, options):
        self.options = options

    def __str__(self):
        return "'%s' is not a valid value for the 'options' parameter" \
               ", please check its structure" % self.options


class BoolOptionsNotValid(OptionsNotValid):
    pass


class IntOptionsNotValid(OptionsNotValid):
    pass


class FloatOptionsNotValid(OptionsNotValid):
    pass
